# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ddt
import mock

from os_faults.ansible import executor
from os_faults.api import node_collection
from os_faults.drivers.cloud import saltcloud
from os_faults.tests.unit import fakes
from os_faults.tests.unit import test


@ddt.ddt
class SaltCloudManagementTestCase(test.TestCase):

    def setUp(self):
        super(SaltCloudManagementTestCase, self).setUp()
        self.fake_ansible_result = fakes.FakeAnsibleResult(
            payload={
                'stdout': 'cmp01.mk20.local:\n'
                          '  eth1:\n'
                          '    hwaddr: 09:7b:74:90:63:c2\n'
                          '    inet:\n'
                          '    - address: 10.0.0.2\n'
                          '  eth2:\n'
                          '    hwaddr: 00:00:00:00:00:02\n'
                          '    inet:\n'
                          '    - address: 192.168.1.2\n'
                          'cmp02.mk20.local:\n'
                          '  eth1:\n'
                          '    hwaddr: 00:00:00:00:00:03\n'
                          '    inet:\n'
                          '    - address: 192.168.1.3\n'
                          '  eth2:\n'
                          '    hwaddr: 09:7b:74:90:63:c3\n'
                          '    inet:\n'
                          '    - address: 10.0.0.3\n'
            })
        self.fake_node_ip_result = fakes.FakeAnsibleResult(
            payload={
                'stdout': 'cmp01.mk20.local:\n'
                          '  10.0.0.2\n'
                          'cmp02.mk20.local:\n'
                          '  10.0.0.3\n'
            })

        self.tcp_conf = {
            'address': 'tcp.local',
            'auth': {
                'username': 'root',
            }
        }
        self.get_nodes_cmd = (
            "salt -E '^(?!cfg|mon)' network.interfaces --out=yaml")
        self.get_ips_cmd = ("salt -E '^(?!cfg|mon)' "
                            "pillar.get _param:single_address --out=yaml")

        self.master_host = node_collection.Host('tcp.local')
        self.hosts = [
            node_collection.Host(ip='10.0.0.2', mac='09:7b:74:90:63:c2',
                                 fqdn='cmp01.mk20.local'),
            node_collection.Host(ip='10.0.0.3', mac='09:7b:74:90:63:c3',
                                 fqdn='cmp02.mk20.local'),
        ]

    @mock.patch('os_faults.ansible.executor.AnsibleRunner', autospec=True)
    @ddt.data((
        dict(address='tcp.local', auth=dict(username='root')),
        (mock.call(auth=dict(username='root')),
         mock.call(auth=dict(jump=dict(username='root', host='tcp.local')),
                   serial=None))
    ), (
        dict(address='tcp.local',
             auth=dict(username='ubuntu',
                       private_key_file='/path/id_rsa'),
             slave_auth=dict(username='root',
                             private_key_file='/path/id_rsa'),
             master_sudo=True),
        (mock.call(auth=dict(username='ubuntu',
                             private_key_file='/path/id_rsa')),
         mock.call(auth=dict(username='root',
                             private_key_file='/path/id_rsa',
                             jump=dict(username='root', host='tcp.local')),
                   serial=None))
    ))
    @ddt.unpack
    def test_init(self, config, expected_runner_calls, mock_ansible_runner):
        ansible_runner_inst = mock_ansible_runner.return_value

        tcp_managment = saltcloud.SaltCloudManagement(config)

        mock_ansible_runner.assert_has_calls(expected_runner_calls)
        self.assertIs(tcp_managment.master_node_executor, ansible_runner_inst)
        self.assertIs(tcp_managment.cloud_executor, ansible_runner_inst)

    @mock.patch('os_faults.ansible.executor.AnsibleRunner', autospec=True)
    def test_verify(self, mock_ansible_runner):
        ansible_runner_inst = mock_ansible_runner.return_value
        ansible_runner_inst.execute.side_effect = [
            [self.fake_ansible_result],
            [self.fake_node_ip_result],
            [fakes.FakeAnsibleResult(payload={'stdout': ''}),
             fakes.FakeAnsibleResult(payload={'stdout': ''})],
        ]
        self.tcp_conf['slave_name_regexp'] = '(ctl*|cmp*)'
        tcp_managment = saltcloud.SaltCloudManagement(self.tcp_conf)
        tcp_managment.verify()

        get_nodes_cmd = "salt -E '(ctl*|cmp*)' network.interfaces --out=yaml"
        get_ips_cmd = ("salt -E '(ctl*|cmp*)' "
                       "pillar.get _param:single_address --out=yaml")
        ansible_runner_inst.execute.assert_has_calls([
            mock.call([self.master_host], {'command': get_nodes_cmd}),
            mock.call([self.master_host], {'command': get_ips_cmd}),
            mock.call(self.hosts, {'command': 'hostname'}),
        ])

    @mock.patch('os_faults.ansible.executor.AnsibleRunner')
    def test_validate_services(self, _):
        tcp_managment = saltcloud.SaltCloudManagement(self.tcp_conf)
        tcp_managment.validate_services()

    @mock.patch('os_faults.ansible.executor.AnsibleRunner', autospec=True)
    def test_get_nodes(self, mock_ansible_runner):
        ansible_runner_inst = mock_ansible_runner.return_value
        ansible_runner_inst.execute.side_effect = [
            [self.fake_ansible_result],
            [self.fake_node_ip_result],
        ]
        tcp_managment = saltcloud.SaltCloudManagement(self.tcp_conf)
        nodes = tcp_managment.get_nodes()

        ansible_runner_inst.execute.assert_has_calls([
            mock.call([self.master_host], {'command': self.get_nodes_cmd}),
            mock.call([self.master_host], {'command': self.get_ips_cmd}),
        ])

        self.assertEqual(nodes.hosts, self.hosts)

    @mock.patch('os_faults.ansible.executor.AnsibleRunner', autospec=True)
    def test_get_nodes_from_discover_driver(self, mock_ansible_runner):
        ansible_runner_inst = mock_ansible_runner.return_value
        hosts = [
            node_collection.Host(ip='10.0.2.2', mac='09:7b:74:90:63:c2',
                                 fqdn='mynode1.local'),
            node_collection.Host(ip='10.0.2.3', mac='09:7b:74:90:63:c3',
                                 fqdn='mynode2.local'),
        ]
        node_discover_driver = mock.Mock()
        node_discover_driver.discover_hosts.return_value = hosts
        tcp_managment = saltcloud.SaltCloudManagement(self.tcp_conf)
        tcp_managment.set_node_discover(node_discover_driver)
        nodes = tcp_managment.get_nodes()

        self.assertFalse(ansible_runner_inst.execute.called)
        self.assertEqual(hosts, nodes.hosts)

    @mock.patch('os_faults.ansible.executor.AnsibleRunner', autospec=True)
    def test_execute_on_cloud(self, mock_ansible_runner):
        ansible_runner_inst = mock_ansible_runner.return_value
        ansible_runner_inst.execute.side_effect = [
            [self.fake_ansible_result],
            [self.fake_node_ip_result],
            [fakes.FakeAnsibleResult(payload={'stdout': ''}),
             fakes.FakeAnsibleResult(payload={'stdout': ''})]
        ]
        tcp_managment = saltcloud.SaltCloudManagement(self.tcp_conf)
        nodes = tcp_managment.get_nodes()
        result = tcp_managment.execute_on_cloud(
            nodes.hosts, {'command': 'mycmd'}, raise_on_error=False)

        ansible_runner_inst.execute.assert_has_calls([
            mock.call([self.master_host], {'command': self.get_nodes_cmd}),
            mock.call([self.master_host], {'command': self.get_ips_cmd}),
            mock.call(self.hosts, {'command': 'mycmd'}, []),
        ])

        self.assertEqual(result,
                         [fakes.FakeAnsibleResult(payload={'stdout': ''}),
                          fakes.FakeAnsibleResult(payload={'stdout': ''})])

    @mock.patch('os_faults.ansible.executor.AnsibleRunner', autospec=True)
    def test_get_nodes_fqdns(self, mock_ansible_runner):
        ansible_runner_inst = mock_ansible_runner.return_value
        ansible_runner_inst.execute.side_effect = [
            [self.fake_ansible_result],
            [self.fake_node_ip_result],
        ]
        tcp_managment = saltcloud.SaltCloudManagement(self.tcp_conf)
        nodes = tcp_managment.get_nodes(fqdns=['cmp02.mk20.local'])

        self.assertEqual(nodes.hosts, [self.hosts[1]])

    @mock.patch('os_faults.ansible.executor.AnsibleRunner', autospec=True)
    @ddt.data(*saltcloud.SaltCloudManagement.SERVICES.keys())
    def test_get_service_nodes(self, service_name, mock_ansible_runner):
        ansible_runner_inst = mock_ansible_runner.return_value
        ansible_runner_inst.execute.side_effect = [
            [self.fake_ansible_result],
            [self.fake_node_ip_result],
            [fakes.FakeAnsibleResult(payload={'stdout': ''},
                                     status=executor.STATUS_FAILED,
                                     host='10.0.0.2'),
             fakes.FakeAnsibleResult(payload={'stdout': ''},
                                     host='10.0.0.3')]
        ]

        tcp_managment = saltcloud.SaltCloudManagement(self.tcp_conf)

        service = tcp_managment.get_service(service_name)
        nodes = service.get_nodes()
        cmd = 'bash -c "ps ax | grep -v grep | grep \'{}\'"'.format(
            service.grep)
        ansible_runner_inst.execute.assert_has_calls([
            mock.call([self.master_host], {'command': self.get_nodes_cmd}),
            mock.call([self.master_host], {'command': self.get_ips_cmd}),
            mock.call(self.hosts, {'command': cmd}, []),
        ])

        self.assertEqual(nodes.hosts, [self.hosts[1]])


@ddt.ddt
class SaltCloudServiceTestCase(test.TestCase):

    def setUp(self):
        super(SaltCloudServiceTestCase, self).setUp()
        self.fake_ansible_result = fakes.FakeAnsibleResult(
            payload={
                'stdout': 'cmp01.mk20.local:\n'
                          '  eth0:\n'
                          '    hwaddr: 09:7b:74:90:63:c2\n'
                          '    inet:\n'
                          '    - address: 10.0.0.2\n'
                          'cmp02.mk20.local:\n'
                          '  eth0:\n'
                          '    hwaddr: 09:7b:74:90:63:c3\n'
                          '    inet:\n'
                          '    - address: 10.0.0.3\n'
            })
        self.fake_node_ip_result = fakes.FakeAnsibleResult(
            payload={
                'stdout': 'cmp01.mk20.local:\n'
                          '  10.0.0.2\n'
                          'cmp02.mk20.local:\n'
                          '  10.0.0.3\n'
            })

        self.tcp_conf = {
            'address': 'tcp.local',
            'auth': {'username': 'root'},
        }
        self.get_nodes_cmd = (
            "salt -E '^(?!cfg|mon)' network.interfaces --out=yaml")
        self.get_ips_cmd = ("salt -E '^(?!cfg|mon)' "
                            "pillar.get _param:single_address --out=yaml")

        self.master_host = node_collection.Host('tcp.local')
        self.hosts = [
            node_collection.Host(ip='10.0.0.2', mac='09:7b:74:90:63:c2',
                                 fqdn='cmp01.mk20.local'),
            node_collection.Host(ip='10.0.0.3', mac='09:7b:74:90:63:c3',
                                 fqdn='cmp02.mk20.local'),
        ]

    @mock.patch('os_faults.ansible.executor.AnsibleRunner', autospec=True)
    @ddt.data(*saltcloud.SaltCloudManagement.SERVICES.keys())
    def test_restart(self, service_name, mock_ansible_runner):
        ansible_runner_inst = mock_ansible_runner.return_value
        ansible_runner_inst.execute.side_effect = [
            [self.fake_ansible_result],
            [self.fake_node_ip_result],
            [fakes.FakeAnsibleResult(payload={'stdout': ''},
                                     status=executor.STATUS_FAILED,
                                     host='10.0.0.2'),
             fakes.FakeAnsibleResult(payload={'stdout': ''},
                                     host='10.0.0.3')],
            [fakes.FakeAnsibleResult(payload={'stdout': ''},
                                     host='10.0.0.2'),
             fakes.FakeAnsibleResult(payload={'stdout': ''},
                                     host='10.0.0.3')]
        ]

        tcp_managment = saltcloud.SaltCloudManagement(self.tcp_conf)

        service = tcp_managment.get_service(service_name)
        service.restart()

        cmd = 'bash -c "ps ax | grep -v grep | grep \'{}\'"'.format(
            service.grep)
        ansible_runner_inst.execute.assert_has_calls([
            mock.call([self.master_host], {'command': self.get_nodes_cmd}),
            mock.call([self.master_host], {'command': self.get_ips_cmd}),
            mock.call(self.hosts,
                      {'command': cmd}, []),
            mock.call([self.hosts[1]], {'shell': service.restart_cmd}),
        ])

    @mock.patch('os_faults.ansible.executor.AnsibleRunner', autospec=True)
    @ddt.data(*saltcloud.SaltCloudManagement.SERVICES.keys())
    def test_terminate(self, service_name, mock_ansible_runner):
        ansible_runner_inst = mock_ansible_runner.return_value
        ansible_runner_inst.execute.side_effect = [
            [self.fake_ansible_result],
            [self.fake_node_ip_result],
            [fakes.FakeAnsibleResult(payload={'stdout': ''},
                                     status=executor.STATUS_FAILED,
                                     host='10.0.0.2'),
             fakes.FakeAnsibleResult(payload={'stdout': ''},
                                     host='10.0.0.3')],
            [fakes.FakeAnsibleResult(payload={'stdout': ''},
                                     host='10.0.0.2'),
             fakes.FakeAnsibleResult(payload={'stdout': ''},
                                     host='10.0.0.3')]
        ]

        tcp_managment = saltcloud.SaltCloudManagement(self.tcp_conf)

        service = tcp_managment.get_service(service_name)
        service.terminate()

        cmd = 'bash -c "ps ax | grep -v grep | grep \'{}\'"'.format(
            service.grep)
        ansible_runner_inst.execute.assert_has_calls([
            mock.call([self.master_host], {'command': self.get_nodes_cmd}),
            mock.call([self.master_host], {'command': self.get_ips_cmd}),
            mock.call(self.hosts,
                      {'command': cmd}, []),
            mock.call([self.hosts[1]], {'shell': service.terminate_cmd}),
        ])

    @mock.patch('os_faults.ansible.executor.AnsibleRunner', autospec=True)
    @ddt.data(*saltcloud.SaltCloudManagement.SERVICES.keys())
    def test_start(self, service_name, mock_ansible_runner):
        ansible_runner_inst = mock_ansible_runner.return_value
        ansible_runner_inst.execute.side_effect = [
            [self.fake_ansible_result],
            [self.fake_node_ip_result],
            [fakes.FakeAnsibleResult(payload={'stdout': ''},
                                     status=executor.STATUS_FAILED,
                                     host='10.0.0.2'),
             fakes.FakeAnsibleResult(payload={'stdout': ''},
                                     host='10.0.0.3')],
            [fakes.FakeAnsibleResult(payload={'stdout': ''},
                                     host='10.0.0.2'),
             fakes.FakeAnsibleResult(payload={'stdout': ''},
                                     host='10.0.0.3')]
        ]

        tcp_managment = saltcloud.SaltCloudManagement(self.tcp_conf)

        service = tcp_managment.get_service(service_name)
        service.start()

        cmd = 'bash -c "ps ax | grep -v grep | grep \'{}\'"'.format(
            service.grep)
        ansible_runner_inst.execute.assert_has_calls([
            mock.call([self.master_host], {'command': self.get_nodes_cmd}),
            mock.call([self.master_host], {'command': self.get_ips_cmd}),
            mock.call(self.hosts,
                      {'command': cmd}, []),
            mock.call([self.hosts[1]], {'shell': service.start_cmd}),
        ])



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging

import os_faults


def main():
    # cloud config schema is an extension to os-client-config
    cloud_config = {
        'cloud_management': {
            'driver': 'fuel',
            'args': {
                'address': 'fuel.local',
                'username': 'root',
            }
        },
        'power_managements': [
            {
                'driver': 'ipmi',
                'args': {
                    'mac_to_bmc': {
                        '00:00:00:00:00:00': {
                            'address': '55.55.55.55',
                            'username': 'foo',
                            'password': 'bar',
                        }
                    }
                }
            }
        ]
    }

    logging.info('Create connection to the cluster')
    cloud_management = os_faults.connect(cloud_config)

    logging.info('Verify connection to the cluster')
    cloud_management.verify()

    logging.info('Get all cluster nodes')
    nodes = cloud_management.get_nodes()
    logging.info('All cluster nodes: %s', nodes)

    computes = nodes.filter(role='compute')

    one = computes.pick()
    logging.info('Pick one of compute nodes: %s', one)

    logging.info('Power off compute node')
    one.poweroff()

    logging.info('Power on compute node')
    one.poweron()

    logging.info('Done!')


if __name__ == '__main__':
    logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s',
                        level=logging.DEBUG)
    main()



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ddt
import mock

from os_faults.api import node_collection
from os_faults.drivers.power import libvirt
from os_faults import error
from os_faults.tests.unit import test


DRIVER_PATH = 'os_faults.drivers.power.libvirt'


@ddt.ddt
class LibvirtDriverTestCase(test.TestCase):

    def setUp(self):
        super(LibvirtDriverTestCase, self).setUp()

        self.params = {'connection_uri': 'fake_connection_uri'}
        self.driver = libvirt.LibvirtDriver(self.params)
        self.host = node_collection.Host(
            ip='10.0.0.2', mac='00:00:00:00:00:00', fqdn='node1.com')

    @mock.patch('oslo_utils.importutils.try_import')
    def test__get_connection_no_cached_connection(self, mock_import):
        mock_libvirt = mock_import.return_value = mock.Mock()
        mock_libvirt_open = mock_libvirt.open = mock.Mock()
        self.driver._get_connection()
        self.assertNotEqual(self.driver._cached_conn, None)

        mock_libvirt_open.assert_called_once_with(
            self.params['connection_uri'])

    def test__get_connection_cached_connection(self):
        self.driver._cached_conn = 'some cached connection'

        conn = self.driver._get_connection()
        self.assertEqual(conn, 'some cached connection')

    @mock.patch(DRIVER_PATH + '.LibvirtDriver._get_connection')
    def test__find_domain_by_host_mac(self, mock__get_connection):
        host = node_collection.Host(ip='10.0.0.2', mac=':54:00:f9:b8:f9')
        domain1 = mock.MagicMock()
        domain1.XMLDesc.return_value = '52:54:00:ab:64:42'
        domain2 = mock.MagicMock()
        domain2.XMLDesc.return_value = '52:54:00:f9:b8:f9'
        self.driver.conn.listAllDomains.return_value = [domain1, domain2]

        domain = self.driver._find_domain_by_host(host)
        self.assertEqual(domain, domain2)

    @mock.patch(DRIVER_PATH + '.LibvirtDriver._get_connection')
    def test__find_domain_by_host_name(self, mock__get_connection):
        host = node_collection.Host(ip='10.0.0.2', libvirt_name='foo')
        domain1 = mock.MagicMock()
        domain1.XMLDesc.return_value = '52:54:00:ab:64:42'
        domain1.name.return_value = 'bar'
        domain2 = mock.MagicMock()
        domain2.XMLDesc.return_value = '52:54:00:f9:b8:f9'
        domain2.name.return_value = 'foo'
        self.driver.conn.listAllDomains.return_value = [domain1, domain2]

        domain = self.driver._find_domain_by_host(host)
        self.assertEqual(domain, domain2)

    @mock.patch(DRIVER_PATH + '.LibvirtDriver._get_connection')
    def test__find_domain_by_host_domain_not_found(
            self, mock__get_connection):
        host = node_collection.Host(ip='10.0.0.2')
        domain1 = mock.MagicMock()
        domain1.XMLDesc.return_value = '52:54:00:ab:64:42'
        domain2 = mock.MagicMock()
        domain2.XMLDesc.return_value = '52:54:00:f9:b8:f9'
        self.driver.conn.listAllDomains.return_value = [domain1, domain2]

        self.assertRaises(error.PowerManagementError,
                          self.driver._find_domain_by_host, host)

    @mock.patch(DRIVER_PATH + '.LibvirtDriver._get_connection')
    def test_supports(self, mock__get_connection):
        domain1 = mock.MagicMock()
        domain1.XMLDesc.return_value = '52:54:00:ab:64:42'
        domain2 = mock.MagicMock()
        domain2.XMLDesc.return_value = '00:00:00:00:00:00'
        self.driver.conn.listAllDomains.return_value = [domain1, domain2]

        self.assertTrue(self.driver.supports(self.host))

    @mock.patch(DRIVER_PATH + '.LibvirtDriver._get_connection')
    def test_supports_false(self, mock__get_connection):
        self.driver.conn.listAllDomains.return_value = []

        self.assertFalse(self.driver.supports(self.host))

    @mock.patch(DRIVER_PATH + '.LibvirtDriver._find_domain_by_host')
    @ddt.data(('poweroff', 'destroy'), ('poweron', 'create'),
              ('reset', 'reset'), ('shutdown', 'shutdown'))
    def test_driver_actions(self, actions, mock__find_domain_by_host):
        getattr(self.driver, actions[0])(self.host)
        domain = mock__find_domain_by_host.return_value
        getattr(domain, actions[1]).assert_called_once_with()

    @mock.patch(DRIVER_PATH + '.LibvirtDriver._find_domain_by_host')
    def test_snapshot(self, mock__find_domain_by_host):
        self.driver.snapshot(self.host, 'foo', suspend=False)
        domain = mock__find_domain_by_host.return_value
        domain.snapshotCreateXML.assert_called_once_with(
            '<domainsnapshot><name>foo</name></domainsnapshot>')

    @mock.patch(DRIVER_PATH + '.LibvirtDriver._find_domain_by_host')
    def test_snapshot_suspend(self, mock__find_domain_by_host):
        self.driver.snapshot(self.host, 'foo', suspend=True)
        domain = mock__find_domain_by_host.return_value
        domain.assert_has_calls((
            mock.call.suspend(),
            mock.call.snapshotCreateXML(
                '<domainsnapshot><name>foo</name></domainsnapshot>'),
            mock.call.resume(),
        ))

    @mock.patch(DRIVER_PATH + '.LibvirtDriver._find_domain_by_host')
    def test_revert(self, mock__find_domain_by_host):
        self.driver.revert(self.host, 'foo', resume=False)
        domain = mock__find_domain_by_host.return_value
        snapshot = domain.snapshotLookupByName.return_value
        domain.snapshotLookupByName.assert_called_once_with('foo')
        domain.revertToSnapshot.assert_called_once_with(snapshot)
        self.assertFalse(domain.resume.called)

    @mock.patch(DRIVER_PATH + '.LibvirtDriver._find_domain_by_host')
    def test_revert_resume(self, mock__find_domain_by_host):
        self.driver.revert(self.host, 'foo', resume=True)
        domain = mock__find_domain_by_host.return_value
        snapshot = domain.snapshotLookupByName.return_value
        domain.snapshotLookupByName.assert_called_once_with('foo')
        domain.revertToSnapshot.assert_called_once_with(snapshot)
        domain.resume.assert_called_once_with()

    @mock.patch(DRIVER_PATH + '.LibvirtDriver._find_domain_by_host')
    def test_revert_destroy(self, mock__find_domain_by_host):
        domain = mock__find_domain_by_host.return_value
        domain.isActive.return_value = True
        self.driver.revert(self.host, 'foo', resume=True)
        domain.destroy.assert_called_once_with()

    @mock.patch(DRIVER_PATH + '.LibvirtDriver._find_domain_by_host')
    def test_revert_destroy_nonactive(self, mock__find_domain_by_host):
        domain = mock__find_domain_by_host.return_value
        domain.isActive.return_value = False
        self.driver.revert(self.host, 'foo', resume=True)
        self.assertFalse(domain.destroy.called)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from os_faults.drivers.services import process

SALT_CALL = 'salt-call --local --retcode-passthrough '
SALT_RESTART = SALT_CALL + 'service.restart {service}'
SALT_TERMINATE = SALT_CALL + 'service.stop {service}'
SALT_START = SALT_CALL + 'service.start {service}'


class SaltService(process.ServiceAsProcess):
    """Salt service

    Service that can be controlled by `salt service.*` commands.

    **Example configuration:**

    .. code-block:: yaml

        services:
          app:
            driver: salt_service
            args:
              salt_service: app
              grep: my_app
              port: ['tcp', 4242]

    parameters:

    - **salt_service** - name of a service
    - **grep** - regexp for grep to find process PID
    - **port** - tuple with two values - protocol, port number (optional)

    """

    NAME = 'salt_service'
    DESCRIPTION = 'Service in salt'
    CONFIG_SCHEMA = {
        'type': 'object',
        'properties': {
            'salt_service': {'type': 'string'},
            'grep': {'type': 'string'},
            'port': process.PORT_SCHEMA,
        },
        'required': ['grep', 'salt_service'],
        'additionalProperties': False,
    }

    def __init__(self, *args, **kwargs):
        super(SaltService, self).__init__(*args, **kwargs)
        self.salt_service = self.config['salt_service']

        self.restart_cmd = SALT_RESTART.format(service=self.salt_service)
        self.terminate_cmd = SALT_TERMINATE.format(service=self.salt_service)
        self.start_cmd = SALT_START.format(service=self.salt_service)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from docutils.parsers import rst
from os_faults import registry
from sphinx.util import docstrings

from ext import utils


class DriverDocDirective(rst.Directive):

    required_arguments = 1

    def run(self):
        drv_name = self.arguments[0]

        driver = registry.get_driver(drv_name)
        types = ', '.join(class_.__name__ for class_ in driver.__bases__)
        doc = '\n'.join(docstrings.prepare_docstring(driver.__doc__))

        subcat = utils.subcategory('{} [{}]'.format(drv_name, types))
        subcat.extend(utils.parse_text(doc))
        return [subcat]


class CloudDriverDocDirective(rst.Directive):

    required_arguments = 1

    def run(self):
        drv_name = self.arguments[0]

        driver = registry.get_driver(drv_name)
        types = ', '.join(class_.__name__ for class_ in driver.__bases__)
        services = sorted(driver.list_supported_services())
        doc = '\n'.join(docstrings.prepare_docstring(driver.__doc__))

        subcat = utils.subcategory('{} [{}]'.format(drv_name, types))
        subcat.extend(utils.parse_text(doc))
        if services:
            subcat.extend(utils.parse_text('**Default services:**'))
            subcat.extend(utils.rstlist(services))
        return [subcat]


def setup(app):
    app.add_directive('driver_doc', DriverDocDirective)
    app.add_directive('cloud_driver_doc', CloudDriverDocDirective)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ddt
import mock

from os_faults.ansible import executor
from os_faults.api import node_collection
from os_faults.tests.unit import test


@ddt.ddt
class AnsibleRunnerTestCase(test.TestCase):

    @mock.patch('os_faults.ansible.executor.os.path.exists')
    def test_resolve_relative_path_doesnt_exist(self, mock_exist):
        mock_exist.return_value = False
        r = executor.resolve_relative_path('')
        self.assertIsNone(r)

    @mock.patch('os_faults.ansible.executor.os.path.exists')
    def test_resolve_relative_path_exists(self, mock_exist):
        mock_exist.return_value = True
        r = executor.resolve_relative_path('')
        self.assertIsNotNone(r)

    @mock.patch('os_faults.ansible.executor.find_ansible')
    @mock.patch.object(executor, 'Options')
    @ddt.data((
        {},
        {},
    ), (
        dict(username='root', password='foobar'),
        dict(ansible_user='root', ansible_ssh_pass='foobar',
             ansible_become_user=None, ansible_become_pass=None,
             ansible_become_method=None,
             ansible_ssh_private_key_file=None,
             ansible_ssh_common_args=executor.SSH_COMMON_ARGS),
    ), (
        dict(username='dev', password='foobar', become_password='secret'),
        dict(ansible_user='dev', ansible_ssh_pass='foobar',
             ansible_become_user=None, ansible_become_pass='secret',
             ansible_become_method=None,
             ansible_ssh_private_key_file=None,
             ansible_ssh_common_args=executor.SSH_COMMON_ARGS),
    ), (
        dict(username='dev', jump={'host': 'jhost.com'},
             private_key_file='/path/my.key'),
        dict(ansible_user='dev', ansible_ssh_pass=None,
             ansible_become_user=None, ansible_become_pass=None,
             ansible_become_method=None,
             ansible_ssh_private_key_file='/path/my.key',
             ansible_ssh_common_args=('-o UserKnownHostsFile=/dev/null '
                                      '-o StrictHostKeyChecking=no '
                                      '-o ConnectTimeout=60 '
                                      '-o ProxyCommand='
                                      '"ssh  '
                                      '-W %h:%p '
                                      '-o UserKnownHostsFile=/dev/null '
                                      '-o StrictHostKeyChecking=no '
                                      '-o ConnectTimeout=60 '
                                      'dev@jhost.com"')),
    ), (
        dict(username='dev', jump={'host': 'jhost.com', 'username': 'juser',
                                   'private_key_file': '/path/my.key'}),
        dict(ansible_user='dev', ansible_ssh_pass=None,
             ansible_become_user=None, ansible_become_pass=None,
             ansible_become_method=None,
             ansible_ssh_private_key_file=None,
             ansible_ssh_common_args=('-o UserKnownHostsFile=/dev/null '
                                      '-o StrictHostKeyChecking=no '
                                      '-o ConnectTimeout=60 '
                                      '-o ProxyCommand='
                                      '"ssh -i /path/my.key '
                                      '-W %h:%p '
                                      '-o UserKnownHostsFile=/dev/null '
                                      '-o StrictHostKeyChecking=no '
                                      '-o ConnectTimeout=60 '
                                      'juser@jhost.com"')),
    ))
    @ddt.unpack
    def test___init__options(self, auth, default_host_vars, mock_options, _):
        runner = executor.AnsibleRunner(auth=auth)
        module_path = executor.make_module_path_option()
        mock_options.assert_called_once_with(
            module_path=module_path, connection='smart', forks=100)
        self.assertEqual(default_host_vars, runner.default_host_vars)

    @mock.patch('os_faults.ansible.executor.find_ansible')
    @mock.patch('os_faults.ansible.executor.AnsibleRunner._run_play')
    def test_run_playbook(self, mock_run_play, _):
        ex = executor.AnsibleRunner()
        my_playbook = [{'gather_facts': 'yes'}, {'gather_facts': 'no'}]
        ex.run_playbook(my_playbook, {})

        self.assertEqual(my_playbook, [{'gather_facts': 'no'},
                                       {'gather_facts': 'no'}])
        self.assertEqual(mock_run_play.call_count, 2)

    @mock.patch('os_faults.ansible.executor.find_ansible')
    @mock.patch('os_faults.ansible.executor.AnsibleRunner.run_playbook')
    def test_execute(self, mock_run_playbook, _):
        my_hosts = [node_collection.Host('0.0.0.0'),
                    node_collection.Host('255.255.255.255')]
        my_tasks = 'my_task'
        ex = executor.AnsibleRunner()
        ex.execute(my_hosts, my_tasks)
        mock_run_playbook.assert_called_once_with(
            [{'tasks': ['my_task'],
              'hosts': ['0.0.0.0', '255.255.255.255'],
              'serial': 10}], {'0.0.0.0': {}, '255.255.255.255': {}})

    @mock.patch('os_faults.ansible.executor.find_ansible')
    @mock.patch('os_faults.ansible.executor.AnsibleRunner.run_playbook')
    def test_execute_with_host_vars(self, mock_run_playbook, _):
        my_hosts = [
            node_collection.Host('0.0.0.0', auth={'username': 'foo',
                                                  'password': 'bar',
                                                  'become_username': 'root'}),
            node_collection.Host('255.255.255.255',
                                 auth={'jump': {'host': '192.168.1.100',
                                                'username': 'foo'}})]
        my_tasks = 'my_task'
        ex = executor.AnsibleRunner()
        ex.execute(my_hosts, my_tasks)
        mock_run_playbook.assert_called_once_with(
            [{'tasks': ['my_task'],
              'hosts': ['0.0.0.0', '255.255.255.255'],
              'serial': 10}],
            {
                '0.0.0.0': {
                    'ansible_user': 'foo',
                    'ansible_ssh_pass': 'bar',
                    'ansible_become_method': None,
                    'ansible_become_user': 'root',
                    'ansible_become_pass': None,
                    'ansible_ssh_private_key_file': None,
                    'ansible_ssh_common_args': executor.SSH_COMMON_ARGS,
                },
                '255.255.255.255': {
                    'ansible_user': None,
                    'ansible_ssh_pass': None,
                    'ansible_become_method': None,
                    'ansible_become_user': None,
                    'ansible_become_pass': None,
                    'ansible_ssh_private_key_file': None,
                    'ansible_ssh_common_args':
                        '-o UserKnownHostsFile=/dev/null '
                        '-o StrictHostKeyChecking=no '
                        '-o ConnectTimeout=60 '
                        '-o ProxyCommand="'
                        'ssh  -W %h:%p '
                        '-o UserKnownHostsFile=/dev/null '
                        '-o StrictHostKeyChecking=no '
                        '-o ConnectTimeout=60 '
                        'foo@192.168.1.100"'}})

    @mock.patch('os_faults.ansible.executor.find_ansible')
    @mock.patch('os_faults.ansible.executor.AnsibleRunner.run_playbook')
    def test_execute_with_serial(self, mock_run_playbook, _):
        my_hosts = [node_collection.Host('0.0.0.0'),
                    node_collection.Host('255.255.255.255')]
        my_tasks = 'my_task'
        ex = executor.AnsibleRunner(serial=50)
        ex.execute(my_hosts, my_tasks)
        mock_run_playbook.assert_called_once_with(
            [{'tasks': ['my_task'],
              'hosts': ['0.0.0.0', '255.255.255.255'],
              'serial': 50}], {'0.0.0.0': {}, '255.255.255.255': {}})

    @mock.patch('os_faults.ansible.executor.find_ansible')
    @mock.patch('os_faults.ansible.executor.AnsibleRunner.run_playbook')
    def test_execute_status_unreachable(self, mock_run_playbook, _):
        my_hosts = [node_collection.Host('0.0.0.0'),
                    node_collection.Host('255.255.255.255')]
        my_tasks = 'my_task'
        my_statuses = {executor.STATUS_FAILED,
                       executor.STATUS_SKIPPED, executor.STATUS_UNREACHABLE}
        r0 = executor.AnsibleExecutionRecord(
            host='0.0.0.0', status=executor.STATUS_OK, task={}, payload={})
        r1 = executor.AnsibleExecutionRecord(
            host='255.255.255.255', status=executor.STATUS_UNREACHABLE,
            task={}, payload={})

        mock_run_playbook.return_value = [r0, r1]
        ex = executor.AnsibleRunner()

        err = self.assertRaises(executor.AnsibleExecutionException,
                                ex.execute, my_hosts, my_tasks, my_statuses)
        self.assertEqual(type(err), executor.AnsibleExecutionUnreachable)

    @mock.patch('os_faults.ansible.executor.find_ansible')
    @mock.patch('os_faults.ansible.executor.AnsibleRunner.run_playbook')
    def test_execute_status_failed(self, mock_run_playbook, _):
        my_hosts = [node_collection.Host('0.0.0.0'),
                    node_collection.Host('255.255.255.255')]
        my_tasks = 'my_task'
        my_statuses = {executor.STATUS_OK, executor.STATUS_FAILED,
                       executor.STATUS_SKIPPED, executor.STATUS_UNREACHABLE}
        r0 = executor.AnsibleExecutionRecord(
            host='0.0.0.0', status=executor.STATUS_OK, task={}, payload={})
        r1 = executor.AnsibleExecutionRecord(
            host='255.255.255.255', status=executor.STATUS_UNREACHABLE,
            task={}, payload={})

        mock_run_playbook.return_value = [r0, r1]
        ex = executor.AnsibleRunner()

        err = self.assertRaises(executor.AnsibleExecutionException,
                                ex.execute, my_hosts, my_tasks, my_statuses)
        self.assertEqual(type(err), executor.AnsibleExecutionException)

    @mock.patch('os_faults.ansible.executor.find_ansible')
    @mock.patch('copy.deepcopy')
    @mock.patch('os_faults.ansible.executor.AnsibleRunner.run_playbook')
    def test_execute_stdout_is_more_than_stdout_limit(
            self, mock_run_playbook, mock_deepcopy, _):
        result = mock.Mock()
        result.payload = {'stdout': 'a' * (executor.STDOUT_LIMIT + 1),
                          'stdout_lines': 'a' * (executor.STDOUT_LIMIT + 1)}
        mock_run_playbook.return_value = [result]

        mock_deepcopy.return_value = [result]
        log_result = mock_deepcopy.return_value[0]

        my_hosts = [node_collection.Host('0.0.0.0'),
                    node_collection.Host('255.255.255.255')]
        my_tasks = 'my_task'
        ex = executor.AnsibleRunner()
        ex.execute(my_hosts, my_tasks)

        self.assertEqual('a' * executor.STDOUT_LIMIT + '... <cut>',
                         log_result.payload['stdout'])

    @mock.patch('os_faults.ansible.executor.find_ansible')
    @mock.patch('os_faults.ansible.executor.LOG.debug')
    @mock.patch('os_faults.ansible.executor.AnsibleRunner.run_playbook')
    def test_execute_payload_without_stdout(self, mock_run_playbook,
                                            mock_debug, _):
        task = {'task': 'foo'}
        host = '0.0.0.0'
        result = executor.AnsibleExecutionRecord(
            host=host, status=executor.STATUS_OK,
            task=task, payload={'foo': 'bar'})
        mock_run_playbook.return_value = [result]

        hosts = [node_collection.Host('0.0.0.0')]
        ex = executor.AnsibleRunner()
        ex.execute(hosts, task)

        mock_debug.assert_has_calls((
            mock.call('Executing task: %s on hosts: %s with serial: %s',
                      task, hosts, 10),
            mock.call('Execution completed with 1 result(s):'),
            mock.call(result),
        ))

    @mock.patch('os_faults.executor.get_module_paths')
    def test_make_module_path_option_ansible_24(self, mock_mp):
        mock_mp.return_value = ['/path/one', 'path/two']
        self.assertEqual(['/path/one', 'path/two'],
                         executor.make_module_path_option())

    @mock.patch('os_faults.executor.get_module_paths')
    def test_make_module_path_option_ansible_24_one_item(self, mock_mp):
        mock_mp.return_value = ['/path/one']
        self.assertEqual(['/path/one', '/path/one'],
                         executor.make_module_path_option())



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import sys

import mock

from os_faults.api import base_driver
from os_faults.api import error
from os_faults import drivers
from os_faults import registry
from os_faults.tests.unit import test


class TestDriver(base_driver.BaseDriver):
    NAME = 'test'


class RegistryTestCase(test.TestCase):

    def setUp(self):
        super(RegistryTestCase, self).setUp()
        registry.DRIVERS.clear()  # reset global drivers list

    def tearDown(self):
        super(RegistryTestCase, self).tearDown()
        registry.DRIVERS.clear()  # reset global drivers list

    @mock.patch('oslo_utils.importutils.import_module')
    @mock.patch('os.walk')
    def test_get_drivers(self, mock_os_walk, mock_import_module):
        drivers_folder = os.path.dirname(drivers.__file__)
        mock_os_walk.return_value = [(drivers_folder, [], ['driver.py'])]
        mock_import_module.return_value = sys.modules[__name__]

        self.assertEqual({'test': TestDriver}, registry.get_drivers())

    @mock.patch('os_faults.registry._list_drivers')
    def test_name_collision(self, mock_list_drivers):
        class TestDriver1(base_driver.BaseDriver):
            NAME = 'test'

        class TestDriver2(base_driver.BaseDriver):
            NAME = 'test'

        mock_list_drivers.return_value = [TestDriver1, TestDriver2]
        self.assertRaises(error.OSFDriverWithSuchNameExists,
                          registry.get_drivers)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ddt
import mock

from os_faults.api import error
from os_faults.drivers.cloud import universal
from os_faults.tests.unit import fakes
from os_faults.tests.unit import test


@ddt.ddt
class UniversalCloudManagementTestCase(test.TestCase):

    @mock.patch('os_faults.ansible.executor.AnsibleRunner', autospec=True)
    @ddt.data((
        dict(),
        dict(),
    ))
    @ddt.unpack
    def test_init(self, config, expected_runner_call, mock_ansible_runner):
        ansible_runner_inst = mock_ansible_runner.return_value

        cloud = universal.UniversalCloudManagement(config)

        mock_ansible_runner.assert_called_with(**expected_runner_call)
        self.assertIs(cloud.cloud_executor, ansible_runner_inst)

    @mock.patch('os_faults.ansible.executor.AnsibleRunner', autospec=True)
    def test_no_discovery(self, mock_ansible_runner):
        address = '10.0.0.10'
        ansible_result = fakes.FakeAnsibleResult(
            payload=dict(stdout='openstack.local'))
        ansible_runner_inst = mock_ansible_runner.return_value
        ansible_runner_inst.execute.side_effect = [
            [ansible_result]
        ]

        cloud = universal.UniversalCloudManagement(dict(address=address))
        self.assertRaises(error.OSFError, cloud.verify)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy

import mock
import six

from os_faults.api import cloud_management
from os_faults.api import error
from os_faults.api import node_collection
from os_faults.api import power_management
from os_faults.tests.unit import test


class MyNodeCollection(node_collection.NodeCollection):
    pass


class NodeCollectionTestCase(test.TestCase):

    def setUp(self):
        super(NodeCollectionTestCase, self).setUp()
        self.mock_cloud_management = mock.Mock(
            spec=cloud_management.CloudManagement)
        self.mock_power_manager = mock.Mock(
            spec=power_management.PowerManager)
        self.mock_cloud_management.power_manager = self.mock_power_manager
        self.hosts = [
            node_collection.Host(ip='10.0.0.2', mac='09:7b:74:90:63:c1',
                                 fqdn='node1.com'),
            node_collection.Host(ip='10.0.0.3', mac='09:7b:74:90:63:c2',
                                 fqdn='node2.com'),
            node_collection.Host(ip='10.0.0.4', mac='09:7b:74:90:63:c3',
                                 fqdn='node3.com'),
            node_collection.Host(ip='10.0.0.5', mac='09:7b:74:90:63:c4',
                                 fqdn='node4.com'),
        ]

        self.node_collection = node_collection.NodeCollection(
            cloud_management=self.mock_cloud_management,
            hosts=copy.deepcopy(self.hosts))

        self.hosts2 = [
            node_collection.Host(ip='10.0.0.7', mac='09:7b:74:90:63:c7',
                                 fqdn='node6.com'),
            node_collection.Host(ip='10.0.0.3', mac='09:7b:74:90:63:c2',
                                 fqdn='node2.com'),
            node_collection.Host(ip='10.0.0.6', mac='09:7b:74:90:63:c6',
                                 fqdn='node5.com'),
            node_collection.Host(ip='10.0.0.2', mac='09:7b:74:90:63:c1',
                                 fqdn='node1.com'),
        ]

        self.node_collection2 = node_collection.NodeCollection(
            cloud_management=self.mock_cloud_management,
            hosts=copy.deepcopy(self.hosts2))

    def test_check_types_wrong_type(self):
        collection = MyNodeCollection(None, [])
        self.assertRaises(TypeError, self.node_collection._check_nodes_types,
                          collection)
        self.assertRaises(TypeError, collection._check_nodes_types,
                          self.node_collection)

    def test_check_types_wrong_cloud_management(self):
        collection = node_collection.NodeCollection(None, [])
        self.assertRaises(error.NodeCollectionError,
                          self.node_collection._check_nodes_types, collection)
        self.assertRaises(error.NodeCollectionError,
                          collection._check_nodes_types, self.node_collection)

    def test_repr(self):
        self.assertIsInstance(repr(self.node_collection), six.string_types)

    def test_len(self):
        self.assertEqual(4, len(self.node_collection))

    def test_add(self):
        collection = self.node_collection + self.node_collection2
        self.assertEqual(['10.0.0.2', '10.0.0.3', '10.0.0.4', '10.0.0.5',
                          '10.0.0.6', '10.0.0.7'],
                         collection.get_ips())

    def test_sub(self):
        collection = self.node_collection - self.node_collection2
        self.assertEqual(['10.0.0.4', '10.0.0.5'],
                         collection.get_ips())

    def test_and(self):
        collection = self.node_collection & self.node_collection2
        self.assertEqual(['10.0.0.2', '10.0.0.3'],
                         collection.get_ips())

    def test_or(self):
        collection = self.node_collection | self.node_collection2
        self.assertEqual(['10.0.0.2', '10.0.0.3', '10.0.0.4', '10.0.0.5',
                          '10.0.0.6', '10.0.0.7'],
                         collection.get_ips())

    def test_xor(self):
        collection = self.node_collection ^ self.node_collection2
        self.assertEqual(['10.0.0.4', '10.0.0.5', '10.0.0.6', '10.0.0.7'],
                         collection.get_ips())

    def test_in(self):
        self.assertIn(self.hosts[0], self.node_collection)

    def test_not_in(self):
        self.assertNotIn(self.hosts2[2], self.node_collection)

    def test_iter(self):
        self.assertEqual(self.hosts, list(self.node_collection))

    def test_getitem(self):
        self.assertEqual(self.hosts[0], self.node_collection[0])

    def test_get_ips(self):
        self.assertEqual(['10.0.0.2', '10.0.0.3', '10.0.0.4', '10.0.0.5'],
                         self.node_collection.get_ips())

    def test_get_macs(self):
        self.assertEqual(['09:7b:74:90:63:c1',
                          '09:7b:74:90:63:c2',
                          '09:7b:74:90:63:c3',
                          '09:7b:74:90:63:c4'],
                         self.node_collection.get_macs())

    def test_get_fqdns(self):
        self.assertEqual(['node1.com', 'node2.com', 'node3.com', 'node4.com'],
                         self.node_collection.get_fqdns())

    def test_pick(self):
        one = self.node_collection.pick()
        self.assertEqual(1, len(one))
        self.assertIn(next(iter(one.hosts)), self.hosts)

    def test_filter(self):
        one = self.node_collection.filter(lambda host: host.ip == '10.0.0.2')
        self.assertEqual(1, len(one))
        self.assertEqual(self.hosts[0], one.hosts[0])

    def test_filter_error(self):
        self.assertRaises(error.NodeCollectionError,
                          self.node_collection.filter,
                          lambda host: host.ip == 'foo')

    def test_run_task(self):
        result = self.node_collection.run_task({'foo': 'bar'},
                                               raise_on_error=False)
        mock_execute_on_cloud = self.mock_cloud_management.execute_on_cloud
        expected_result = mock_execute_on_cloud.return_value
        self.assertIs(result, expected_result)
        mock_execute_on_cloud.assert_called_once_with(
            self.hosts, {'foo': 'bar'}, raise_on_error=False)

    def test_pick_count(self):
        two = self.node_collection.pick(count=2)
        self.assertEqual(2, len(two))
        for host in two:
            self.assertIn(host, self.node_collection)

    def test_pick_exception(self):
        self.assertRaises(
            error.NodeCollectionError, self.node_collection.pick, count=10)

    def test_poweroff(self):
        self.node_collection.poweroff()
        self.mock_power_manager.poweroff.assert_called_once_with(self.hosts)

    def test_poweron(self):
        self.node_collection.poweron()
        self.mock_power_manager.poweron.assert_called_once_with(self.hosts)

    def test_reset(self):
        self.node_collection.reset()
        self.mock_power_manager.reset.assert_called_once_with(self.hosts)

    def test_shutdown(self):
        self.node_collection.shutdown()
        self.mock_power_manager.shutdown.assert_called_once_with(self.hosts)

    def test_reboot(self):
        self.node_collection.reboot()
        self.mock_cloud_management.execute_on_cloud.assert_called_once_with(
            self.hosts, {'command': 'reboot now'})

    def test_snapshot(self):
        self.node_collection.snapshot('foo')
        self.mock_power_manager.snapshot.assert_called_once_with(
            self.hosts, 'foo', True)

    def test_revert(self):
        self.node_collection.revert('foo')
        self.mock_power_manager.revert.assert_called_once_with(
            self.hosts, 'foo', True)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
import os
import sys

from oslo_utils import importutils

import os_faults
from os_faults.api import base_driver
from os_faults.api import error

DRIVERS = {}


def _import_modules_from_package():
    folder = os.path.dirname(os_faults.__file__)
    library_root = os.path.normpath(os.path.join(folder, os.pardir))
    drivers_folder = os.path.join(folder, 'drivers')

    for root, dirs, files in os.walk(drivers_folder):
        for filename in files:
            if (filename.startswith('__') or
                    filename.startswith('test') or
                    not filename.endswith('.py')):
                continue

            relative_path = os.path.relpath(os.path.join(root, filename),
                                            library_root)
            name = os.path.splitext(relative_path)[0]  # remove extension
            module_name = '.'.join(name.split(os.sep))  # convert / to .

            if module_name not in sys.modules:
                module = importutils.import_module(module_name)
                sys.modules[module_name] = module
            else:
                module = sys.modules[module_name]

            yield module


def _list_drivers():
    modules = _import_modules_from_package()

    for module in modules:
        class_info_list = inspect.getmembers(module, inspect.isclass)

        for class_info in class_info_list:
            klazz = class_info[1]
            if not issubclass(klazz, base_driver.BaseDriver):
                continue
            if 'NAME' not in vars(klazz):  # driver must have a name
                continue
            if klazz.NAME == 'base':  # skip base class
                continue

            yield klazz


def get_drivers():
    global DRIVERS

    if not DRIVERS:
        DRIVERS = {}
        for k in _list_drivers():
            driver_name = k.get_driver_name()
            if driver_name in DRIVERS:
                orig_k = DRIVERS[driver_name]
                orig_path = orig_k.__module__ + '.' + orig_k.__name__
                dup_path = k.__module__ + '.' + k.__name__

                raise error.OSFDriverWithSuchNameExists(
                    'Driver "%s" already defined in %s. '
                    'Found a duplicate in %s ' % (
                        driver_name, orig_path, dup_path))
            DRIVERS[driver_name] = k

    return DRIVERS


def get_driver(name):
    all_drivers = get_drivers()

    if name not in all_drivers:
        raise error.OSFDriverNotFound('Driver %s is not found' % name)

    return all_drivers[name]



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from os_faults.ansible import executor
from os_faults.api import cloud_management
from os_faults.api import node_collection
from os_faults.api import node_discover
from os_faults.drivers import shared_schemas

LOG = logging.getLogger(__name__)


class DevStackNodeCollection(node_collection.NodeCollection):
    def connect(self, network_name):
        raise NotImplementedError

    def disconnect(self, network_name):
        raise NotImplementedError


class DevStackCloudManagement(cloud_management.CloudManagement,
                              node_discover.NodeDiscover):
    """Driver for DevStack.

    This driver requires DevStack installed with Systemd (USE_SCREEN=False).
    Supports discovering of node MAC addresses.

    **Example configuration:**

    .. code-block:: yaml

        cloud_management:
          driver: devstack
          args:
            address: 192.168.1.10
            auth:
              username: ubuntu
              password: ubuntu_pass
              private_key_file: ~/.ssh/id_rsa_devstack
            iface: eth1

    parameters:

    - **address** - ip address of any devstack node
    - **username** - username for all nodes
    - **password** - password for all nodes (optional)
    - **private_key_file** - path to key file (optional)
    - **iface** - network interface name to retrieve mac address (optional)
    """

    NAME = 'devstack'
    DESCRIPTION = 'DevStack driver'
    NODE_CLS = DevStackNodeCollection
    SERVICES = {
        'cinder-api': {
            'driver': 'system_service',
            'args': {
                'grep': 'cinder-api',
                'service_name': 'devstack@c-api',
            }
        },
        'cinder-scheduler': {
            'driver': 'system_service',
            'args': {
                'grep': 'cinder-schedule',
                'service_name': 'devstack@c-sch',
            }
        },
        'cinder-volume': {
            'driver': 'system_service',
            'args': {
                'grep': 'cinder-volume',
                'service_name': 'devstack@c-vol',
            }
        },
        'etcd': {
            'driver': 'system_service',
            'args': {
                'grep': 'etcd',
                'service_name': 'devstack@etcd',
            }
        },
        'glance-api': {
            'driver': 'system_service',
            'args': {
                'grep': 'glance-api',
                'service_name': 'devstack@g-api',
            }
        },
        'heat-api': {
            'driver': 'system_service',
            'args': {
                'grep': 'heat-api',
                'service_name': 'devstack@h-api',
            }
        },
        'heat-engine': {
            'driver': 'system_service',
            'args': {
                'grep': 'heat-engine',
                'service_name': 'devstack@h-eng',
            }
        },
        'keystone': {
            'driver': 'system_service',
            'args': {
                'grep': 'keystone',
                'service_name': 'devstack@keystone',
            }
        },
        'memcached': {
            'driver': 'system_service',
            'args': {
                'grep': 'memcached',
                'service_name': 'memcached',
            }
        },
        'mysql': {
            'driver': 'system_service',
            'args': {
                'grep': 'mysqld',
                'service_name': 'mariadb',
                'port': ['tcp', 3307],
            }
        },
        'neutron-dhcp-agent': {
            'driver': 'system_service',
            'args': {
                'grep': 'neutron-dhcp-agent',
                'service_name': 'devstack@q-dhcp',
            }
        },
        'neutron-l3-agent': {
            'driver': 'system_service',
            'args': {
                'grep': 'neutron-l3-agent',
                'service_name': 'devstack@q-l3',
            }
        },
        'neutron-meta-agent': {
            'driver': 'system_service',
            'args': {
                'grep': 'neutron-meta-agent',
                'service_name': 'devstack@q-meta',
            }
        },
        'neutron-openvswitch-agent': {
            'driver': 'system_service',
            'args': {
                'grep': 'neutron-openvswitch-agent',
                'service_name': 'devstack@q-agt',
            }
        },
        'neutron-server': {
            'driver': 'system_service',
            'args': {
                'grep': 'neutron-server',
                'service_name': 'devstack@q-svc',
            }
        },
        'nova-api': {
            'driver': 'system_service',
            'args': {
                'grep': 'nova-api',
                'service_name': 'devstack@n-api',
            }
        },
        'nova-compute': {
            'driver': 'system_service',
            'args': {
                'grep': 'nova-compute',
                'service_name': 'devstack@n-cpu',
            }
        },
        'nova-scheduler': {
            'driver': 'system_service',
            'args': {
                'grep': 'nova-scheduler',
                'service_name': 'devstack@n-sch',
            }
        },
        'placement-api': {
            'driver': 'system_service',
            'args': {
                'grep': 'placement',
                'service_name': 'devstack@placement-api',
            }
        },
        'rabbitmq': {
            'driver': 'system_service',
            'args': {
                'grep': 'rabbitmq_server',
                'service_name': 'rabbitmq-server',
            }
        },
    }
    SUPPORTED_NETWORKS = ['all-in-one']
    CONFIG_SCHEMA = {
        'type': 'object',
        '$schema': 'http://json-schema.org/draft-04/schema#',
        'properties': {
            'address': {'type': 'string'},
            'auth': shared_schemas.AUTH_SCHEMA,
            'iface': {'type': 'string'},
        },
        'required': ['address', 'auth'],
        'additionalProperties': False,
    }

    def __init__(self, cloud_management_params):
        super(DevStackCloudManagement, self).__init__()
        self.node_discover = self  # supports discovering

        address = cloud_management_params['address']
        auth = cloud_management_params['auth']
        self.iface = cloud_management_params.get('iface', 'eth0')

        self.cloud_executor = executor.AnsibleRunner(auth=auth)

        self.hosts = [node_collection.Host(ip=address)]
        self.nodes = None

    def verify(self):
        """Verify connection to the cloud."""
        nodes = self.get_nodes()
        if nodes:
            LOG.debug('DevStack nodes: %s', nodes)
            LOG.info('Connected to cloud successfully')

    def execute_on_cloud(self, hosts, task, raise_on_error=True):
        """Execute task on specified hosts within the cloud.

        :param hosts: List of host FQDNs
        :param task: Ansible task
        :param raise_on_error: throw exception in case of error
        :return: Ansible execution result (list of records)
        """
        if raise_on_error:
            return self.cloud_executor.execute(hosts, task)
        else:
            return self.cloud_executor.execute(hosts, task, [])

    def discover_hosts(self):
        if self.nodes is None:
            get_mac_cmd = 'cat /sys/class/net/{}/address'.format(self.iface)
            task = {'command': get_mac_cmd}
            results = self.execute_on_cloud(self.hosts, task)

            # TODO(astudenov): support fqdn
            self.nodes = [node_collection.Host(ip=r.host,
                                               mac=r.payload['stdout'],
                                               fqdn='')
                          for r in results]

        return self.nodes



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from os_faults.api import node_collection
from os_faults.drivers.nodes import node_list
from os_faults.tests.unit import test


class NodeListDiscoverTestCase(test.TestCase):

    def test_discover_hosts(self):
        conf = [
            {'ip': '10.0.0.11', 'mac': '01', 'fqdn': 'node-1'},
            {'ip': '10.0.0.12', 'mac': '02', 'fqdn': 'node-2'},
        ]
        expected_hosts = [
            node_collection.Host(ip='10.0.0.11', mac='01', fqdn='node-1'),
            node_collection.Host(ip='10.0.0.12', mac='02', fqdn='node-2'),
        ]

        node_list_discover = node_list.NodeListDiscover(conf)
        hosts = node_list_discover.discover_hosts()
        self.assertEqual(expected_hosts, hosts)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import os
import re
import shlex

from oslo_concurrency import processutils

from os_faults.tests.unit import test


LOG = logging.getLogger(__name__)


class TestOSInjectFaultUniversalDriver(test.TestCase):
    CONFIG_FILE = os.path.join(os.path.dirname(__file__),
                               'os-faults-universal.yaml')

    def test_connect(self):
        cmd = 'os-inject-fault -dc %s -v' % self.CONFIG_FILE

        command_stdout, command_stderr = processutils.execute(
            *shlex.split(cmd))

        success = re.search('Connected to cloud successfully', command_stderr)
        self.assertTrue(success)

    def test_restart_etcd(self):
        cmd = 'os-inject-fault -dc %s restart etcd service' % self.CONFIG_FILE

        command_stdout, command_stderr = processutils.execute(
            *shlex.split(cmd))

        print('stdout', command_stdout)
        print('stderr', command_stderr)
        success = re.search('successfully', command_stderr)
        self.assertTrue(success)


class TestOSInjectFaultDevstackDriver(TestOSInjectFaultUniversalDriver):
    CONFIG_FILE = os.path.join(os.path.dirname(__file__),
                               'os-faults-devstack.yaml')



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import argparse
import inspect
import logging
import sys
import textwrap

import os_faults
from os_faults.api import cloud_management
from os_faults.api import node_collection as node_collection_pkg
from os_faults.api import service as service_pkg
from os_faults import registry


def describe_actions(klazz):
    methods = (m for m in inspect.getmembers(
        klazz,
        predicate=lambda o: ((inspect.isfunction(o) or inspect.ismethod(o)) and
                             hasattr(o, '__public__'))))
    return ['%s - %s' % (m[0], m[1].__doc__.split('\n')[0])
            for m in sorted(methods)]

SERVICE_ACTIONS = describe_actions(service_pkg.Service)
NODE_ACTIONS = describe_actions(node_collection_pkg.NodeCollection)

USAGE = 'os-inject-fault [-h] [-c CONFIG] [-d] [-v] [command]'
HELP_TEMPLATE = """
Built-in drivers:
%(drivers)s

*Service-oriented* commands perform specified action against service on
all, on one random node or on the node specified by FQDN:

  <action> <service> service [on (random|one|single|<fqdn> node[s])]

  where:
    action is one of:
      %(service_action)s
    service is one of supported by driver:
%(services)s

Examples:
    * "Restart Keystone service" - restarts Keystone service on all nodes.
    * "kill nova-api service on one node" - restarts Nova API on one
      randomly-picked node.

*Node-oriented* commands perform specified action on node specified by FQDN
or set of service's nodes:

  <action> [random|one|single|<fqdn>] node[s] [with <service> service]

  where:
    action is one of:
      %(node_action)s
    service is one of supported by driver:
%(services)s

Examples:
    * "Reboot one node with mysql" - reboots one random node with MySQL.
    * "Reset node-2.domain.tld node" - reset node node-2.domain.tld.

*Network-oriented* commands are subset of node-oriented and perform network
management operation on selected nodes:

  [connect|disconnect] <network> network on [random|one|single|<fqdn>] node[s]
    [with <service> service]

  where:
    network is one of supported by driver:
%(networks)s
    service is one of supported by driver:
%(services)s

Examples:
    * "Disconnect management network on nodes with rabbitmq service" - shuts
      down management network interface on all nodes where rabbitmq runs.
    * "Connect storage network on node-1.domain.tld node" - enables storage
      network interface on node-1.domain.tld.

For more details please refer to docs: http://os-faults.readthedocs.io/
"""


def _list_items(group, items):
    s = '\n'.join(
        textwrap.wrap(', '.join(sorted(items)),
                      subsequent_indent=' ' * (len(group) + 8),
                      break_on_hyphens=False))
    if not s:
        s = '/no built-in support/'
    return '      %s: %s' % (group, s)


def _make_epilog():
    drivers = registry.get_drivers()
    services_strings = []
    networks_strings = []
    driver_descriptions = []

    for driver_name, driver in sorted(drivers.items(), key=lambda x: x[0]):
        driver_descriptions.append(
            '  %s - %s' % (driver_name, driver.get_driver_description()))

        if issubclass(driver, cloud_management.CloudManagement):
            services_strings.append(
                _list_items(driver_name, driver.list_supported_services()))
            networks_strings.append(
                _list_items(driver_name, driver.list_supported_networks()))

    return HELP_TEMPLATE % dict(
        drivers='\n'.join(driver_descriptions),
        service_action='\n      '.join(SERVICE_ACTIONS),
        services='\n'.join(services_strings),
        node_action='\n      '.join(sorted(NODE_ACTIONS)),
        networks='\n'.join(networks_strings),
    )


def main():
    parser = argparse.ArgumentParser(
        prog='os-inject-fault',
        usage=USAGE,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog=_make_epilog())
    parser.add_argument('-c', '--config', dest='config',
                        help='path to os-faults cloud connection config')
    parser.add_argument('-d', '--debug', dest='debug', action='store_true')
    parser.add_argument('-v', '--verify', action='store_true',
                        help='verify connection to the cloud')
    parser.add_argument('command', nargs='*',
                        help='fault injection command, e.g. "restart keystone '
                             'service"')
    args = parser.parse_args()

    debug = args.debug
    logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s',
                        level=logging.DEBUG if debug else logging.INFO)

    config = args.config
    command = args.command

    if not command and not args.verify:
        parser.print_help()
        sys.exit(0)

    destructor = os_faults.connect(config_filename=config)

    if args.verify:
        destructor.verify()

    if command:
        command = ' '.join(command)
        os_faults.human_api(destructor, command)

if __name__ == '__main__':
    main()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging

import os_faults


def main():
    cloud_config = {
        'cloud_management': {
            'driver': 'universal',
        },
        'node_discover': {
            'driver': 'node_list',
            'args': [
                {
                    'ip': '192.168.5.127',
                    'auth': {
                        'username': 'root',
                        'private_key_file': 'openstack_key',
                    }
                },
                {
                    'ip': '192.168.5.128',
                    'auth': {
                        'username': 'root',
                        'private_key_file': 'openstack_key',
                    }
                }
            ]
        },
        'services': {
            'memcached': {
                'driver': 'system_service',
                'args': {
                    'service_name': 'memcached',
                    'grep': 'memcached',
                }
            }
        },
        'power_managements': [
            {
                'driver': 'libvirt',
                'args': {
                    'connection_uri': 'qemu+unix:///system',
                }
            },
        ]
    }

    logging.info('# Create connection to the cloud')
    cloud_management = os_faults.connect(cloud_config)

    logging.info('# Verify connection to the cloud')
    cloud_management.verify()

    logging.info('# Kill Memcached service on all nodes')
    service = cloud_management.get_service(name='memcached')
    service.kill()


if __name__ == '__main__':
    logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s',
                        level=logging.INFO)
    main()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging

import os_faults


def main():
    # cloud config schema is an extension to os-client-config
    cloud_config = {
        'cloud_management': {
            'driver': 'devstack',
            'args': {
                'address': 'devstack.local',
                'username': 'stack',
            }
        }
    }

    logging.info('# Create connection to the cloud')
    cloud_management = os_faults.connect(cloud_config)

    logging.info('# Verify connection to the cloud')
    cloud_management.verify()

    logging.info('# Restart Keystone service on all nodes')
    service = cloud_management.get_service(name='keystone')
    service.restart()


if __name__ == '__main__':
    logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s',
                        level=logging.INFO)
    main()



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc

import six

from os_faults.api import base_driver
from os_faults.api import error
from os_faults import utils


@six.add_metaclass(abc.ABCMeta)
class PowerDriver(base_driver.BaseDriver):

    @abc.abstractmethod
    def supports(host):
        """Returns True if host is supported by the power driver"""

    @abc.abstractmethod
    def poweroff(self, host):
        """Power off host abruptly"""

    @abc.abstractmethod
    def poweron(self, host):
        """Power on host"""

    @abc.abstractmethod
    def reset(self, host):
        """Reset host"""

    @abc.abstractmethod
    def shutdown(self, host):
        """Graceful shutdown host"""

    def snapshot(self, host, snapshot_name, suspend=True):
        raise NotImplementedError

    def revert(self, host, snapshot_name, resume=True):
        raise NotImplementedError


class PowerManager(object):

    def __init__(self):
        self.power_drivers = []

    def add_driver(self, driver):
        self.power_drivers.append(driver)

    def _map_hosts_to_driver(self, hosts):
        driver_host_pairs = []
        for host in hosts:
            for power_driver in self.power_drivers:
                if power_driver.supports(host):
                    driver_host_pairs.append((power_driver, host))
                    break
            else:
                raise error.PowerManagementError(
                    "No supported driver found for host {}".format(host))
        return driver_host_pairs

    def _run_command(self, cmd, hosts, **kwargs):
        driver_host_pairs = self._map_hosts_to_driver(hosts)
        tw = utils.ThreadsWrapper()
        for driver, host in driver_host_pairs:
            kwargs['host'] = host
            fn = getattr(driver, cmd)
            tw.start_thread(fn, **kwargs)
        tw.join_threads()
        if tw.errors:
            raise error.PowerManagementError(
                'There are some errors when working the driver. '
                'Please, check logs for more details.')

    def poweroff(self, hosts):
        self._run_command('poweroff', hosts)

    def poweron(self, hosts):
        self._run_command('poweron', hosts)

    def reset(self, hosts):
        self._run_command('reset', hosts)

    def shutdown(self, hosts):
        self._run_command('shutdown', hosts)

    def snapshot(self, hosts, snapshot_name, suspend=True):
        self._run_command('snapshot', hosts,
                          snapshot_name=snapshot_name, suspend=suspend)

    def revert(self, hosts, snapshot_name, resume=True):
        self._run_command('revert', hosts,
                          snapshot_name=snapshot_name, resume=resume)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import logging
import threading

LOG = logging.getLogger(__name__)

MACADDR_REGEXP = '^([0-9A-Fa-f]{2}:){5}[0-9A-Fa-f]{2}$'
FQDN_REGEXP = '.*'


class ThreadsWrapper(object):
    def __init__(self):
        self.threads = []
        self.errors = []

    def _target(self, fn, **kwargs):
        try:
            fn(**kwargs)
        except Exception as exc:
            LOG.error('%s raised exception: %s', fn, exc)
            self.errors.append(exc)

    def start_thread(self, fn, **kwargs):
        thread = threading.Thread(target=self._target,
                                  args=(fn, ), kwargs=kwargs)
        thread.start()
        self.threads.append(thread)

    def join_threads(self):
        for thread in self.threads:
            thread.join()


def require_variables(*variables):
    """Class method decorator to check that required variables are present"""
    def decorator(fn):
        @functools.wraps(fn)
        def wrapper(self, *args, **kawrgs):
            missing_vars = []
            for var in variables:
                if not getattr(self, var, None):
                    missing_vars.append(var)
            if missing_vars:
                missing_vars = ', '.join(missing_vars)
                msg = '{} required for {}.{}'.format(
                    missing_vars, self.__class__.__name__, fn.__name__)
                raise NotImplementedError(msg)
            return fn(self, *args, **kawrgs)
        return wrapper
    return decorator


class ComparableMixin(object):

    ATTRS = ()

    def _cmp_attrs(self):
        return tuple(getattr(self, attr) for attr in self.ATTRS)

    def __lt__(self, other):
        return self._cmp_attrs() < other._cmp_attrs()

    def __le__(self, other):
        return self._cmp_attrs() <= other._cmp_attrs()

    def __eq__(self, other):
        return self._cmp_attrs() == other._cmp_attrs()

    def __ge__(self, other):
        return self._cmp_attrs() >= other._cmp_attrs()

    def __gt__(self, other):
        return self._cmp_attrs() > other._cmp_attrs()

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self._cmp_attrs())


class ReprMixin(object):

    ATTRS = ()
    REPR_ATTRS = ()

    def __repr__(self):
        return '{}({})'.format(
            self.__class__.__name__,
            ', '.join('{}={}'.format(attr, repr(getattr(self, attr)))
                      for attr in self.REPR_ATTRS or self.ATTRS))



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from os_faults.api import node_collection
from os_faults.api import node_discover
from os_faults import utils


AUTH_SCHEMA = {
    'type': 'object',
    'properties': {
        'username': {'type': 'string'},
        'password': {'type': 'string'},
        'private_key_file': {'type': 'string'},
        'become_password': {'type': 'string'},
        'jump': {
            'type': 'object',
            'properties': {
                'host': {'type': 'string'},
                'username': {'type': 'string'},
                'private_key_file': {'type': 'string'},
            },
            'required': ['host'],
            'additionalProperties': False,
        },
    },
    'additionalProperties': False,
}


class NodeListDiscover(node_discover.NodeDiscover):
    """Node list.

    Allows specifying list of nodes in configuration.

    **Example configuration:**

    .. code-block:: yaml

        node_discover:
          driver: node_list
          args:
          - ip: 10.0.0.51
            mac: aa:bb:cc:dd:ee:01
            fqdn: node1.local
            libvirt_name: node1
          - ip: 192.168.1.50
            mac: aa:bb:cc:dd:ee:02
            fqdn: node2.local
            auth:
              username: user1
              password: secret1
              jump:
                host: 10.0.0.52
                username: ubuntu
                private_key_file: /path/to/file
          - ip: 10.0.0.53
            mac: aa:bb:cc:dd:ee:03
            fqdn: node3.local
            become_password: my_secret_password

    node parameters:

    - **ip** - ip/host of the node
    - **mac** - MAC address of the node (optional).
      MAC address is used for libvirt driver.
    - **fqdn** - FQDN of the node (optional).
      FQDN is used for filtering only.
    - **auth** - SSH related parameters (optional):
        - **username** - SSH username (optional)
        - **password** - SSH password (optional)
        - **private_key_file** - SSH key file (optional)
        - **become_password** - privilege escalation password (optional)
        - **jump** - SSH proxy parameters (optional):
            - **host** - SSH proxy host
            - **username** - SSH proxy user
            - **private_key_file** - SSH proxy key file (optional)

    """

    NAME = 'node_list'
    DESCRIPTION = 'Reads hosts from configuration file'
    CONFIG_SCHEMA = {
        '$schema': 'http://json-schema.org/draft-04/schema#',
        'type': 'array',
        'items': {
            'type': 'object',
            'properties': {
                'ip': {'type': 'string'},
                'mac': {'type': 'string', 'pattern': utils.MACADDR_REGEXP},
                'fqdn': {'type': 'string'},
                'libvirt_name': {'type': 'string'},
                'auth': AUTH_SCHEMA,
            },
            'required': ['ip'],
            'additionalProperties': False,
        },
        'minItems': 1,
    }

    def __init__(self, conf):
        self.hosts = [node_collection.Host(**host) for host in conf]

    def discover_hosts(self):
        """Discover hosts

        :returns: list of Host instances
        """
        return self.hosts



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class OSFException(Exception):
    """Base Exception class"""


class OSFError(OSFException):
    """Base Error class"""


class PowerManagementError(OSFError):
    """Base Error class for Power Management API"""


class CloudManagementError(OSFError):
    """Base Error class for Cloud Management API"""


class ServiceError(OSFError):
    """Base Error class for Service API"""


class ContainerError(OSFError):
    """Base Error class for Container API"""


class NodeCollectionError(OSFError):
    """Base Error class for NodeCollection API"""


class OSFDriverNotFound(OSFError):
    """Driver Not Found by os-faults registry"""


class OSFDriverWithSuchNameExists(OSFError):
    """Driver with such name already exists in os-faults registry"""
